##############################################################################
# Copyright 2016-2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Iterator, Optional, List

import rpcq
from qcs_api_client.client import QCSClientConfiguration
from rpcq.messages import TargetDevice as TargetQuantumProcessor


@dataclass
class CompileToNativeQuilRequest:
    """
    Request to compile to native Quil.
    """

    program: str
    """Program to compile."""

    target_quantum_processor: TargetQuantumProcessor
    """Quantum processor to target."""

    protoquil: Optional[bool]
    """Whether or not to restrict to protoquil. Overrides server default when provided."""


@dataclass
class NativeQuilMetadataResponse:
    """
    Metadata for a native Quil program.
    """

    final_rewiring: List[int]
    """Output qubit index relabeling due to SWAP insertion."""

    gate_depth: Optional[int]
    """Maximum number of successive gates in the native Quil program."""

    gate_volume: Optional[int]
    """Total number of gates in the native Quil program."""

    multiqubit_gate_depth: Optional[int]
    """Maximum number of successive two-qubit gates in the native Quil program."""

    program_duration: Optional[float]
    """Rough estimate of native Quil program length in nanoseconds."""

    program_fidelity: Optional[float]
    """Rough estimate of the fidelity of the full native Quil program."""

    topological_swaps: Optional[int]
    """Total number of SWAPs in the native Quil program."""

    qpu_runtime_estimation: Optional[float]
    """
    The estimated runtime (milliseconds) on a Rigetti QPU (protoquil program). Available only for protoquil-compliant
    programs.
    """


@dataclass
class CompileToNativeQuilResponse:
    """
    Compile to native Quil response.
    """

    native_program: str
    """Native Quil program."""

    metadata: Optional[NativeQuilMetadataResponse]
    """Metadata for the returned Native Quil."""


@dataclass
class ConjugatePauliByCliffordRequest:
    """
    Request to conjugate a Pauli element by a Clifford element.
    """

    pauli_indices: List[int]
    """Qubit indices onto which the factors of the Pauli term are applied."""

    pauli_symbols: List[str]
    """Ordered factors of the Pauli term."""

    clifford: str
    """Clifford element."""


@dataclass
class ConjugatePauliByCliffordResponse:
    """
    Conjugate Pauli by Clifford response.
    """

    phase_factor: int
    """Encoded global phase factor on the emitted Pauli."""

    pauli: str
    """Description of the encoded Pauli."""


@dataclass
class GenerateRandomizedBenchmarkingSequenceRequest:
    """
    Request to generate a randomized benchmarking sequence.
    """

    depth: int
    """Depth of the benchmarking sequence."""

    num_qubits: int
    """Number of qubits involved in the benchmarking sequence."""

    gateset: List[str]
    """List of Quil programs, each describing a Clifford."""

    seed: Optional[int]
    """PRNG seed. Set this to guarantee repeatable results."""

    interleaver: Optional[str]
    """Fixed Clifford, specified as a Quil string, to interleave through an RB sequence."""


@dataclass
class GenerateRandomizedBenchmarkingSequenceResponse:
    """
    Randomly generated benchmarking sequence response.
    """

    sequence: List[List[int]]
    """List of Cliffords, each expressed as a list of generator indices."""


class CompilerClient:
    """
    Client for making requests to a Quil compiler.
    """

    def __init__(self, *, client_configuration: QCSClientConfiguration, request_timeout: float = 10.0) -> None:
        """
        Instantiate a new compiler client.

        :param client_configuration: Configuration for client.
        :param request_timeout: Timeout for requests, in seconds.
        """
        base_url = client_configuration.profile.applications.pyquil.quilc_url
        if not base_url.startswith("tcp://"):
            raise ValueError(f"Expected compiler URL '{base_url}' to start with 'tcp://'")

        self.base_url = base_url
        self.timeout = request_timeout

    def get_version(self) -> str:
        """
        Get version info for compiler server.
        """
        with self._rpcq_client() as rpcq_client:  # type: rpcq.Client
            version: Optional[str] = rpcq_client.call("get_version_info").get("quilc")
            if version is None:
                raise ValueError("Expected compiler version info to contain a 'quilc' field.")

            return version

    def compile_to_native_quil(self, request: CompileToNativeQuilRequest) -> CompileToNativeQuilResponse:
        """
        Compile Quil program to native Quil.
        """
        rpcq_request = rpcq.messages.NativeQuilRequest(
            quil=request.program,
            target_device=request.target_quantum_processor,
        )
        with self._rpcq_client() as rpcq_client:  # type: rpcq.Client
            response: rpcq.messages.NativeQuilResponse = rpcq_client.call(
                "quil_to_native_quil",
                rpcq_request,
                protoquil=request.protoquil,
            )
            metadata: Optional[NativeQuilMetadataResponse] = None
            if response.metadata is not None:
                metadata = NativeQuilMetadataResponse(
                    final_rewiring=response.metadata.final_rewiring,
                    gate_depth=response.metadata.gate_depth,
                    gate_volume=response.metadata.gate_volume,
                    multiqubit_gate_depth=response.metadata.multiqubit_gate_depth,
                    program_duration=response.metadata.program_duration,
                    program_fidelity=response.metadata.program_fidelity,
                    topological_swaps=response.metadata.topological_swaps,
                    qpu_runtime_estimation=response.metadata.qpu_runtime_estimation,
                )
            return CompileToNativeQuilResponse(native_program=response.quil, metadata=metadata)

    def conjugate_pauli_by_clifford(self, request: ConjugatePauliByCliffordRequest) -> ConjugatePauliByCliffordResponse:
        """
        Conjugate a Pauli element by a Clifford element.
        """
        rpcq_request = rpcq.messages.ConjugateByCliffordRequest(
            pauli=rpcq.messages.PauliTerm(indices=request.pauli_indices, symbols=request.pauli_symbols),
            clifford=request.clifford,
        )
        with self._rpcq_client() as rpcq_client:  # type: rpcq.Client
            response: rpcq.messages.ConjugateByCliffordResponse = rpcq_client.call(
                "conjugate_pauli_by_clifford",
                rpcq_request,
            )
            return ConjugatePauliByCliffordResponse(phase_factor=response.phase, pauli=response.pauli)

    def generate_randomized_benchmarking_sequence(
        self, request: GenerateRandomizedBenchmarkingSequenceRequest
    ) -> GenerateRandomizedBenchmarkingSequenceResponse:
        """
        Generate a randomized benchmarking sequence.
        """
        rpcq_request = rpcq.messages.RandomizedBenchmarkingRequest(
            depth=request.depth,
            qubits=request.num_qubits,
            gateset=request.gateset,
            seed=request.seed,
            interleaver=request.interleaver,
        )
        with self._rpcq_client() as rpcq_client:  # type: rpcq.Client
            response: rpcq.messages.RandomizedBenchmarkingResponse = rpcq_client.call(
                "generate_rb_sequence",
                rpcq_request,
            )
            return GenerateRandomizedBenchmarkingSequenceResponse(sequence=response.sequence)

    @contextmanager
    def _rpcq_client(self) -> Iterator[rpcq.Client]:
        client = rpcq.Client(
            endpoint=self.base_url,
            timeout=self.timeout,
        )
        try:
            yield client
        finally:
            client.close()  # type: ignore



##############################################################################
# Copyright 2016-2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import threading
from datetime import datetime
from typing import Dict, NamedTuple, Optional, TYPE_CHECKING

from dateutil.parser import parse as parsedate
from dateutil.tz import tzutc
from qcs_api_client.client import QCSClientConfiguration
from qcs_api_client.models import EngagementWithCredentials, CreateEngagementRequest
from qcs_api_client.operations.sync import create_engagement
from qcs_api_client.types import UNSET
from qcs_api_client.util.errors import QCSHTTPStatusError

from pyquil.api._qcs_client import qcs_client

if TYPE_CHECKING:
    import httpx


class QPUUnavailableError(Exception):
    """
    Exception raised when a QPU is unavailable.
    """

    retry_after: Optional[int]
    """The number of seconds after which to retry the engagement request."""

    def __init__(self, retry_after: Optional[str]) -> None:
        if retry_after is not None:
            super().__init__(f"QPU unavailable. Please retry after {retry_after}s.")
        else:
            super().__init__("QPU unavailable.")


class EngagementCacheKey(NamedTuple):
    quantum_processor_id: str
    endpoint_id: Optional[str]


class EngagementManager:
    """
    Fetches (and caches) engagements for use when accessing a QPU.
    """

    _lock: threading.Lock
    """Lock used to ensure that only one engagement request is in flight at once."""

    def __init__(self, *, client_configuration: QCSClientConfiguration) -> None:
        """
        Instantiate a new engagement manager.

        :param client_configuration: Client configuration, used for refreshing engagements.
        """
        self._client_configuration = client_configuration
        self._cached_engagements: Dict[EngagementCacheKey, EngagementWithCredentials] = {}
        self._lock = threading.Lock()

    def get_engagement(
        self, *, quantum_processor_id: str, request_timeout: float = 10.0, endpoint_id: Optional[str] = None
    ) -> EngagementWithCredentials:
        """
        Gets an engagement for the given quantum processor endpoint.

        If an engagement was already fetched previously and remains valid, it will be returned instead
        of creating a new engagement.

        :param quantum_processor_id: Quantum processor being engaged.
        :param request_timeout: Timeout for request, in seconds.
        :param endpoint_id: Optional ID of the endpoint to use for engagement. If provided, it must
            correspond to an endpoint serving the provided Quantum Processor.
        :return: Fetched or cached engagement.
        :raises QPUUnavailableError: raised when the QPU is unavailable due, and provides a suggested
            number of seconds to wait until retrying.
        """
        key = EngagementCacheKey(quantum_processor_id, endpoint_id)

        with self._lock:
            if not self._engagement_valid(self._cached_engagements.get(key)):
                with qcs_client(
                    client_configuration=self._client_configuration, request_timeout=request_timeout
                ) as client:  # type: httpx.Client
                    request = CreateEngagementRequest(
                        quantum_processor_id=quantum_processor_id, endpoint_id=endpoint_id or UNSET
                    )
                    response = create_engagement(client=client, json_body=request)
                    try:
                        self._cached_engagements[key] = response.parsed
                    except QCSHTTPStatusError as e:
                        if response.status_code == 503:
                            raise QPUUnavailableError(retry_after=response.headers.get("Retry-After")) from e
            return self._cached_engagements[key]

    @staticmethod
    def _engagement_valid(engagement: Optional[EngagementWithCredentials]) -> bool:
        if engagement is None:
            return False

        return all(
            [
                engagement.credentials.client_public != "",
                engagement.credentials.client_secret != "",
                engagement.credentials.server_public != "",
                parsedate(engagement.expires_at) > datetime.now(tzutc()),
                engagement.address != "",
            ]
        )



from dataclasses import dataclass, field
import dataclasses
from typing import Dict, Mapping, Sequence, Union

from rpcq.messages import ParameterAref

ParameterValue = Union[int, float, Sequence[int], Sequence[float]]


@dataclass
class Memory:
    """
    Memory encapsulates the values to be sent as parameters alongside a program at time of
    execution, and read back afterwards.
    """

    values: Dict[ParameterAref, Union[int, float]] = field(default_factory=dict)

    def copy(self) -> "Memory":
        """
        Return a deep copy of this Memory object.
        """
        return Memory(values={dataclasses.replace(k): v for k, v in self.values.items()})

    def write(self, parameter_values: Mapping[Union[str, ParameterAref], ParameterValue]) -> "Memory":
        """
        Set the given values for the given parameters.
        """
        for parameter, parameter_value in parameter_values.items():
            self._write_value(parameter=parameter, value=parameter_value)
        return self

    def _write_value(
        self,
        *,
        parameter: Union[ParameterAref, str],
        value: ParameterValue,
    ) -> "Memory":
        """
        Mutate the program to set the given parameter value.

        :param parameter: Name of the memory region, or parameter reference with offset.
        :param value: the value or values to set for this parameter. If a list
        is provided, parameter must be a ``str`` or ``parameter.offset == 0``.
        """
        if isinstance(parameter, str):
            parameter = ParameterAref(name=parameter, index=0)

        import numpy as np

        if isinstance(value, (int, float)):
            self.values[parameter] = value
        elif isinstance(value, (Sequence, np.ndarray)):
            if parameter.index != 0:
                raise ValueError("Parameter may not have a non-zero index when its value is a sequence")

            for index, v in enumerate(value):
                if not isinstance(v, (int, float)):
                    raise TypeError(f"Parameter must be numeric, not {type(value)}")
                aref = ParameterAref(name=parameter.name, index=index)
                self.values[aref] = v
        else:
            raise TypeError(f"Parameter must be numeric or an iterable of numeric values, not {type(value)}")

        return self



from typing import Dict, List, Union, Optional, Any
from typing_extensions import Literal
from pydantic import BaseModel, Field
from rpcq.messages import TargetDevice as TargetQuantumProcessor


class Operator(BaseModel):
    operator: Optional[str] = None
    duration: Optional[float] = None
    fidelity: Optional[float] = None


class MeasureInfo(Operator):
    qubit: Optional[Union[int, str]] = None
    target: Optional[Union[int, str]] = None
    operator_type: Literal["measure"] = "measure"


class GateInfo(Operator):
    parameters: List[Union[float, str]] = Field(default_factory=list)
    arguments: List[Union[int, str]] = Field(default_factory=list)
    operator_type: Literal["gate"] = "gate"


class Qubit(BaseModel):
    id: int
    dead: Optional[bool] = False
    gates: List[Union[GateInfo, MeasureInfo]] = Field(default_factory=list)

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        exclude = kwargs.get("exclude") or set()
        if not self.dead:
            exclude.add("dead")
        kwargs["exclude"] = exclude
        return super().dict(**kwargs)


class Edge(BaseModel):
    ids: List[int]
    dead: Optional[bool] = False
    gates: List[GateInfo] = Field(default_factory=list)

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        exclude = kwargs.get("exclude") or set()
        if not self.dead:
            exclude.add("dead")
        kwargs["exclude"] = exclude
        return super().dict(**kwargs)


class CompilerISA(BaseModel):
    qubits: Dict[str, Qubit] = Field(default_factory=dict, alias="1Q")
    edges: Dict[str, Edge] = Field(default_factory=dict, alias="2Q")


def add_qubit(quantum_processor: CompilerISA, node_id: int) -> Qubit:
    if node_id not in quantum_processor.qubits:
        quantum_processor.qubits[str(node_id)] = Qubit(id=node_id)
    return quantum_processor.qubits[str(node_id)]


def get_qubit(quantum_processor: CompilerISA, node_id: int) -> Optional[Qubit]:
    return quantum_processor.qubits.get(str(node_id))


def make_edge_id(qubit1: int, qubit2: int) -> str:
    return "-".join([str(qubit) for qubit in sorted([qubit1, qubit2])])


def add_edge(quantum_processor: CompilerISA, qubit1: int, qubit2: int) -> Edge:
    edge_id = make_edge_id(qubit1, qubit2)
    if edge_id not in quantum_processor.edges:
        quantum_processor.edges[edge_id] = Edge(ids=sorted([qubit1, qubit2]))
    return quantum_processor.edges[edge_id]


def get_edge(quantum_processor: CompilerISA, qubit1: int, qubit2: int) -> Optional[Edge]:
    edge_id = make_edge_id(qubit1, qubit2)
    return quantum_processor.edges.get(edge_id)


def _edge_ids_from_id(edge_id: str) -> List[int]:
    return [int(node_id) for node_id in edge_id.split("-")]


def _compiler_isa_from_dict(data: Dict[str, Dict[str, Any]]) -> CompilerISA:
    compiler_isa_data = {
        "1Q": {k: {"id": int(k), **v} for k, v in data.get("1Q", {}).items()},
        "2Q": {k: {"ids": _edge_ids_from_id(k), **v} for k, v in data.get("2Q", {}).items()},
    }
    return CompilerISA.parse_obj(compiler_isa_data)


def compiler_isa_to_target_quantum_processor(compiler_isa: CompilerISA) -> TargetQuantumProcessor:
    return TargetQuantumProcessor(isa=compiler_isa.dict(by_alias=True), specs={})


class Supported1QGate:
    I = "I"
    RX = "RX"
    RZ = "RZ"
    MEASURE = "MEASURE"
    WILDCARD = "WILDCARD"


class Supported2QGate:
    WILDCARD = "WILDCARD"
    CZ = "CZ"
    ISWAP = "ISWAP"
    CPHASE = "CPHASE"
    XY = "XY"



import json
import pkgutil
import operator
from typing import List

from lark import Lark, Transformer, v_args

import numpy as np

from pyquil.quilbase import (
    AbstractInstruction,
    DefGate,
    DefPermutationGate,
    DefGateByPaulis,
    DefWaveform,
    Qubit,
    FormalArgument,
    Frame,
    Pulse,
    Fence,
    FenceAll,
    DefCalibration,
    DefMeasureCalibration,
    DefFrame,
    Parameter,
    Declare,
    Capture,
    RawCapture,
    MemoryReference,
    Pragma,
    RawInstr,
    JumpTarget,
    Jump,
    JumpWhen,
    JumpUnless,
    Reset,
    ResetQubit,
    Wait,
    ClassicalStore,
    ClassicalLoad,
    ClassicalConvert,
    ClassicalExchange,
    ClassicalMove,
    ClassicalNeg,
    ClassicalNot,
    ClassicalAnd,
    ClassicalInclusiveOr,
    ClassicalExclusiveOr,
    ClassicalAdd,
    ClassicalSub,
    ClassicalMul,
    ClassicalDiv,
    ClassicalEqual,
    ClassicalGreaterEqual,
    ClassicalGreaterThan,
    ClassicalLessThan,
    ClassicalLessEqual,
)
from pyquil.quiltwaveforms import _wf_from_dict
from pyquil.quilatom import (
    WaveformReference,
    Expression,
    quil_sqrt,
    quil_sin,
    quil_cos,
    quil_cis,
    quil_exp,
    Label,
    _contained_mrefs,
)
from pyquil.gates import (
    DELAY,
    SHIFT_PHASE,
    SET_PHASE,
    SWAP_PHASE,
    SET_SCALE,
    SET_FREQUENCY,
    SHIFT_FREQUENCY,
    QUANTUM_GATES,
    MEASURE,
    HALT,
    NOP,
    Gate,
)


class QuilTransformer(Transformer):  # type: ignore
    def quil(self, instructions):
        return instructions

    indented_instrs = list

    @v_args(inline=True)
    def def_gate_matrix(self, name, variables, matrix):
        return DefGate(name, matrix=matrix, parameters=variables)

    @v_args(inline=True)
    def def_gate_as_permutation(self, name, matrix):
        return DefPermutationGate(name, permutation=matrix)

    @v_args(inline=True)
    def def_pauli_gate(self, name, variables, qubits, terms):
        pg = DefGateByPaulis(name, parameters=variables, arguments=qubits, body=terms)
        return pg

    pauli_terms = list

    @v_args(inline=True)
    def pauli_term(self, name, expression, qubits):
        from pyquil.paulis import PauliTerm

        return PauliTerm.from_list(list(zip(name, qubits)), expression)

    @v_args(inline=True)
    def def_circuit(self, name, variables, qubits, instrs):
        qubits = qubits if qubits else []
        space = " " if qubits else ""
        if variables:
            raw_defcircuit = "DEFCIRCUIT {}({}){}{}:".format(
                name, ", ".join(map(str, variables)), space, " ".join(map(str, qubits))
            )
        else:
            raw_defcircuit = "DEFCIRCUIT {}{}{}:".format(name, space, " ".join(map(str, qubits)))

        raw_defcircuit += "\n    ".join([""] + [str(instr) for instr in instrs])
        return RawInstr(raw_defcircuit)

    @v_args(inline=True)
    def def_circuit_without_qubits(self, name, variables, instrs):
        return self.def_circuit_qubits(name, variables, [], instrs)

    @v_args(inline=True)
    def def_frame(self, frame, *specs):
        names = {
            "DIRECTION": "direction",
            "HARDWARE-OBJECT": "hardware_object",
            "INITIAL-FREQUENCY": "initial_frequency",
            "SAMPLE-RATE": "sample_rate",
            "CENTER-FREQUENCY": "center_frequency",
        }
        options = {}

        for (spec_name, spec_value) in specs:
            name = names.get(spec_name, None)
            if name:
                options[name] = json.loads(str(spec_value))
            else:
                raise ValueError(
                    f"Unexpectected attribute {spec_name} in definition of frame {frame}. " f"{frame}, {specs}"
                )

        f = DefFrame(frame, **options)
        return f

    frame_spec = list
    frame_attr = v_args(inline=True)(str)

    @v_args(inline=True)
    def def_waveform(self, name, params, matrix):
        return DefWaveform(name, params, matrix[0])

    @v_args(inline=True)
    def def_calibration(self, name, params, qubits, instructions):
        for p in params:
            mrefs = _contained_mrefs(p)
            if mrefs:
                raise ValueError(f"Unexpected memory references {mrefs} in DEFCAL {name}. Did you forget a '%'?")
        dc = DefCalibration(name, params, qubits, instructions)
        return dc

    @v_args(inline=True)
    def def_measure_calibration(self, qubit, name, instructions):
        mref = FormalArgument(name) if name else None
        dmc = DefMeasureCalibration(qubit, mref, instructions)
        return dmc

    @v_args(inline=True)
    def gate(self, modifiers, name, params, qubits):
        # TODO Don't like this.
        modifiers = modifiers or []
        params = params or []

        # Some gate modifiers increase the arity of the base gate. The
        # new qubit arguments prefix the old ones.
        modifier_qubits = []
        for m in modifiers:
            if m in ["CONTROLLED", "FORKED"]:
                modifier_qubits.append(qubits[len(modifier_qubits)])

        base_qubits = qubits[len(modifier_qubits) :]
        forked_offset = len(params) >> modifiers.count("FORKED")
        base_params = params[:forked_offset]

        if name in QUANTUM_GATES:
            if base_params:
                gate = QUANTUM_GATES[name](*base_params, *base_qubits)
            else:
                gate = QUANTUM_GATES[name](*base_qubits)
        else:
            gate = Gate(name, base_params, base_qubits)

        for modifier in modifiers[::-1]:
            if modifier == "CONTROLLED":
                gate.controlled(modifier_qubits.pop())
            elif modifier == "DAGGER":
                gate.dagger()
            elif modifier == "FORKED":
                gate.forked(modifier_qubits.pop(), params[forked_offset : (2 * forked_offset)])
                forked_offset *= 2
            else:
                raise ValueError(f"Unsupported gate modifier {modifier}.")

        return gate

    @v_args(inline=True)
    def gate_no_qubits(self, name):
        return RawInstr(name)

    modifiers = list
    modifier = v_args(inline=True)(str)

    @v_args(inline=True)
    def frame(self, qubits, name):
        f = Frame(qubits, name)
        return f

    @v_args(inline=True)
    def pulse(self, nonblocking, frame, waveform):
        p = Pulse(frame, waveform, nonblocking=bool(nonblocking))
        return p

    @v_args(inline=True)
    def fence_some(self, qubits):
        f = Fence(list(qubits))
        return f

    fence_all = v_args(inline=True)(FenceAll)

    @v_args(inline=True)
    def declare(self, name, memory_type, memory_size, *sharing):
        shared, *offsets = sharing
        d = Declare(
            str(name),
            memory_type=str(memory_type),
            memory_size=int(memory_size) if memory_size else 1,
            shared_region=str(shared) if shared else None,
            offsets=offsets if shared else None,
        )
        return d

    @v_args(inline=True)
    def capture(self, nonblocking, frame, waveform, addr):
        c = Capture(frame, waveform, addr, nonblocking=nonblocking)
        return c

    @v_args(inline=True)
    def raw_capture(self, nonblocking, frame, expression, addr):
        c = RawCapture(frame, expression, addr, nonblocking=nonblocking)
        return c

    @v_args(inline=True)
    def addr(self, name):
        return MemoryReference(str(name))

    @v_args(inline=True)
    def addr_subscript(self, name, subscript):
        return MemoryReference(str(name), int(subscript))

    @v_args(inline=True)
    def offset_descriptor(self, offset, name):
        return (int(offset), str(name))

    @v_args(inline=True)
    def delay_qubits(self, qubits, delay_amount=None):
        # TODO(notmgsk): This is a very nasty hack. I can't quite get
        # the Lark grammar to recognize the last token (i.e. 1) in
        # `DELAY 0 1` as the delay amount. I think it's because it
        # matches 1 as a qubit rather than an expression (in the
        # grammar). Then again I would expect look-ahead to see that
        # it matches expression too, so it should give that
        # preference. How do we encode that priority?
        if delay_amount is None:
            delay_amount = int(qubits[-1].index)
            qubits = qubits[:-1]
        d = DELAY(*[*qubits, delay_amount])
        return d

    @v_args(inline=True)
    def delay_frames(self, qubit, *frames_and_delay_amount):
        *frame_names, delay_amount = frames_and_delay_amount
        frames = [Frame([qubit], name) for name in frame_names]
        d = DELAY(*[*frames, delay_amount])
        return d

    @v_args(inline=True)
    def shift_phase(self, frame, expression):
        return SHIFT_PHASE(frame, expression)

    @v_args(inline=True)
    def set_phase(self, frame, expression):
        return SET_PHASE(frame, expression)

    @v_args(inline=True)
    def set_scale(self, frame, expression):
        return SET_SCALE(frame, expression)

    @v_args(inline=True)
    def set_frequency(self, frame, expression):
        return SET_FREQUENCY(frame, expression)

    @v_args(inline=True)
    def shift_frequency(self, frame, expression):
        return SHIFT_FREQUENCY(frame, expression)

    @v_args(inline=True)
    def swap_phase(self, framea, frameb):
        return SWAP_PHASE(framea, frameb)

    @v_args(inline=True)
    def pragma(self, name, *pragma_names_and_string):
        args = list(map(str, pragma_names_and_string))
        p = Pragma(str(name), args=args)
        return p

    @v_args(inline=True)
    def pragma_freeform_string(self, name, *pragma_names_and_string):
        if len(pragma_names_and_string) == 1:
            freeform_string = pragma_names_and_string[0]
            args = ()
        else:
            *args_identifiers, freeform_string = pragma_names_and_string
            args = list(map(str, args_identifiers))
        # Strip the quotes from start/end of string which are matched
        # by the Lark grammar
        freeform_string = freeform_string[1:-1]
        p = Pragma(str(name), args=args, freeform_string=freeform_string)
        return p

    @v_args(inline=True)
    def measure(self, qubit, address):
        return MEASURE(qubit, address)

    @v_args(inline=True)
    def halt(self):
        return HALT

    @v_args(inline=True)
    def nop(self):
        return NOP

    @v_args(inline=True)
    def include(self, string):
        return RawInstr(f"INCLUDE {string}")

    @v_args(inline=True)
    def def_label(self, label):
        return JumpTarget(label)

    @v_args(inline=True)
    def jump(self, label):
        return Jump(label)

    @v_args(inline=True)
    def jump_when(self, label, address):
        return JumpWhen(label, address)

    @v_args(inline=True)
    def jump_unless(self, label, address):
        return JumpUnless(label, address)

    label = v_args(inline=True)(Label)

    @v_args(inline=True)
    def reset(self, qubit):
        if qubit:
            return ResetQubit(qubit)
        else:
            return Reset()

    @v_args(inline=True)
    def wait(self):
        return Wait()

    @v_args(inline=True)
    def store(self, left, subscript, right):
        return ClassicalStore(left, subscript, right)

    @v_args(inline=True)
    def load(self, left, right, subscript):
        return ClassicalLoad(left, right, subscript)

    @v_args(inline=True)
    def convert(self, left, right):
        return ClassicalConvert(left, right)

    @v_args(inline=True)
    def exchange(self, left, right):
        return ClassicalExchange(left, right)

    @v_args(inline=True)
    def move(self, left, right):
        return ClassicalMove(left, right)

    @v_args(inline=True)
    def classical_unary(self, op, target):
        if op == "TRUE":
            return ClassicalMove(target, 1)
        elif op == "FALSE":
            return ClassicalMove(target, 0)
        elif op == "NEG":
            return ClassicalNeg(target)
        elif op == "NOT":
            return ClassicalNot(target)

    @v_args(inline=True)
    def logical_binary_op(self, op, left, right):
        if op == "AND":
            return ClassicalAnd(left, right)
        elif op == "OR":
            return ClassicalInclusiveOr(left, right)
        elif op == "IOR":
            return ClassicalInclusiveOr(left, right)
        elif op == "XOR":
            return ClassicalExclusiveOr(left, right)

    @v_args(inline=True)
    def arithmetic_binary_op(self, op, left, right):
        if op == "ADD":
            return ClassicalAdd(left, right)
        elif op == "SUB":
            return ClassicalSub(left, right)
        elif op == "MUL":
            return ClassicalMul(left, right)
        elif op == "DIV":
            return ClassicalDiv(left, right)

    @v_args(inline=True)
    def classical_comparison(self, op, target, left, right):
        if op == "EQ":
            return ClassicalEqual(target, left, right)
        elif op == "GT":
            return ClassicalGreaterThan(target, left, right)
        elif op == "GE":
            return ClassicalGreaterEqual(target, left, right)
        elif op == "LT":
            return ClassicalLessThan(target, left, right)
        elif op == "LE":
            return ClassicalLessEqual(target, left, right)

    @v_args(inline=True)
    def waveform(self, name, *params):
        param_dict = {k: v for (k, v) in params}
        if param_dict:
            return _wf_from_dict(name, param_dict)
        else:
            return WaveformReference(name)

    @v_args(inline=True)
    def waveform_name(self, prefix, suffix=None):
        return f"{prefix}/{suffix}" if suffix else prefix

    def matrix(self, rows):
        return list(rows)

    def matrix_row(self, expressions):
        return list(expressions)

    def params(self, params):
        return list(params)

    @v_args(inline=True)
    def named_param(self, name, val):
        return (str(name), val)

    def qubit_designators(self, qubits):
        return list(qubits)

    qubit = v_args(inline=True)(Qubit)
    qubits = list
    qubit_variable = v_args(inline=True)(FormalArgument)
    qubit_variables = list

    @v_args(inline=True)
    def variable(self, var):
        variable = Parameter(str(var))
        return variable

    def variables(self, variables):
        return list(variables)

    @v_args(inline=True)
    def i(self):
        return 1j

    @v_args(inline=True)
    def imag(self, number):
        return number * 1j

    @v_args(inline=True)
    def pi(self):
        return np.pi

    int_n = v_args(inline=True)(int)
    float_n = v_args(inline=True)(float)

    name = v_args(inline=True)(str)
    string = v_args(inline=True)(str)

    @v_args(inline=True)
    def signed_number(self, sign, number):
        if sign and sign == "-":
            return -number
        else:
            return number

    @v_args(inline=True)
    def apply_fun(self, fun, arg):
        if fun == "SIN":
            return quil_sin(arg) if isinstance(arg, Expression) else np.sin(arg)
        if fun == "COS":
            return quil_cos(arg) if isinstance(arg, Expression) else np.cos(arg)
        if fun == "SQRT":
            return quil_sqrt(arg) if isinstance(arg, Expression) else np.sqrt(arg)
        if fun == "EXP":
            return quil_exp(arg) if isinstance(arg, Expression) else np.exp(arg)
        if fun == "CIS":
            return quil_cis(arg) if isinstance(arg, Expression) else np.cos(arg) + 1j * np.sin(arg)

    add = v_args(inline=True)(operator.add)
    sub = v_args(inline=True)(operator.sub)
    mul = v_args(inline=True)(operator.mul)
    div = v_args(inline=True)(operator.truediv)
    pow = v_args(inline=True)(operator.pow)
    neg = v_args(inline=True)(operator.neg)
    pos = v_args(inline=True)(operator.pos)
    function = v_args(inline=True)(str)
    keyword = v_args(inline=True)(str)


grammar = pkgutil.get_data("pyquil._parser", "grammar.lark").decode()
parser = Lark(
    grammar,
    start="quil",
    parser="lalr",
    transformer=QuilTransformer(),
    maybe_placeholders=True,
)


def run_parser(program: str) -> List[AbstractInstruction]:
    """
    Parse a raw Quil program and return a corresponding list of PyQuil objects.

    :param str quil: a single or multiline Quil program
    :return: list of instructions
    """
    p = parser.parse(program)
    return p



__all__ = [
    "get_measure_probabilities",
    "NumpyWavefunctionSimulator",
    "ReferenceDensitySimulator",
    "ReferenceWavefunctionSimulator",
    "targeted_einsum",
    "targeted_tensordot",
    "zero_state_matrix",
]

from pyquil.simulation._numpy import (
    NumpyWavefunctionSimulator,
    get_measure_probabilities,
    targeted_einsum,
    targeted_tensordot,
)
from pyquil.simulation._reference import (
    ReferenceDensitySimulator,
    ReferenceWavefunctionSimulator,
    zero_state_matrix,
)



from copy import copy
from dataclasses import dataclass
from numbers import Complex, Real
from typing import Callable, Dict, Union, List, Optional, no_type_check

import numpy as np
from scipy.special import erf

from pyquil.quilatom import TemplateWaveform, _update_envelope, _complex_str, Expression, substitute

_waveform_classes: Dict[str, type] = {}
"""A mapping from Quil-T waveform names to their corresponding classes.

This should not be mutated directly, but rather filled by the @waveform
decorator.
"""


def waveform(name: str) -> Callable[[type], type]:
    """ Define a Quil-T wavefom with the given name. """

    def wrap(cls: type) -> type:
        cls: type = dataclass(cls)
        _waveform_classes[name] = cls
        return cls

    return wrap


@no_type_check
def _wf_from_dict(name: str, params: Dict[str, Union[Expression, Real, Complex]]) -> TemplateWaveform:
    """Construct a TemplateWaveform from a name and a dictionary of properties.

    :param name: The Quil-T name of the template.
    :param params: A mapping from parameter names to their corresponding values.
    :returns: A template waveform.
    """
    params = copy(params)
    if name not in _waveform_classes:
        raise ValueError(f"Unknown template waveform {name}.")
    cls = _waveform_classes[name]
    fields = getattr(cls, "__dataclass_fields__", {})

    for param, value in params.items():
        if param not in fields:
            raise ValueError(f"Unexpected parameter '{param}' in {name}.")

        if isinstance(value, Expression):
            value = substitute(value, {})

        if isinstance(value, Real):
            # normalize to float
            params[param] = float(value)
        elif isinstance(value, Complex):
            # no normalization needed
            pass
        else:
            raise ValueError(f"Unable to resolve parameter '{param}' in template {name} to a constant value.")

    for field, spec in fields.items():
        if field not in params and spec.default is not None:
            raise ValueError(f"Missing parameter '{field}' in {name}.")

    return cls(**params)


def _optional_field_strs(wf: TemplateWaveform) -> List[str]:
    """Get the printed representations of optional template parameters."""
    result = []
    for field, spec in getattr(wf, "__dataclass_fields__", {}).items():
        if spec.default is None:
            value = getattr(wf, field, None)
            if value is not None:
                result.append(f"{field}: {value}")
    return result


@waveform("flat")
class FlatWaveform(TemplateWaveform):
    """
    A flat (constant) waveform.
    """

    iq: Complex
    """ A raw IQ value. """

    scale: Optional[float] = None
    """ An optional global scaling factor. """

    phase: Optional[float] = None
    """ An optional phase shift factor. """

    detuning: Optional[float] = None
    """ An optional frequency detuning factor. """

    def out(self) -> str:
        output = "flat("
        output += ", ".join([f"duration: {self.duration}", f"iq: {_complex_str(self.iq)}"] + _optional_field_strs(self))
        output += ")"
        return output

    def __str__(self) -> str:
        return self.out()

    def samples(self, rate: float) -> np.ndarray:
        iqs = np.full(self.num_samples(rate), self.iq, dtype=np.complex128)
        return _update_envelope(iqs, rate, scale=self.scale, phase=self.phase, detuning=self.detuning)


@waveform("gaussian")
class GaussianWaveform(TemplateWaveform):
    """ A Gaussian pulse. """

    fwhm: float
    """ The Full-Width-Half-Max of the Gaussian (seconds). """

    t0: float
    """ The center time coordinate of the Gaussian (seconds). """

    scale: Optional[float] = None
    """ An optional global scaling factor. """

    phase: Optional[float] = None
    """ An optional phase shift factor. """

    detuning: Optional[float] = None
    """ An optional frequency detuning factor. """

    def out(self) -> str:
        output = "gaussian("
        output += ", ".join(
            [f"duration: {self.duration}", f"fwhm: {self.fwhm}", f"t0: {self.t0}"] + _optional_field_strs(self)
        )
        output += ")"
        return output

    def __str__(self) -> str:
        return self.out()

    def samples(self, rate: float) -> np.ndarray:
        ts = np.arange(self.num_samples(rate), dtype=np.complex128) / rate
        sigma = 0.5 * self.fwhm / np.sqrt(2.0 * np.log(2.0))
        iqs = np.exp(-0.5 * (ts - self.t0) ** 2 / sigma ** 2)
        return _update_envelope(iqs, rate, scale=self.scale, phase=self.phase, detuning=self.detuning)


@waveform("drag_gaussian")
class DragGaussianWaveform(TemplateWaveform):
    """ A DRAG Gaussian pulse. """

    fwhm: float
    """ The Full-Width-Half-Max of the gaussian (seconds). """

    t0: float
    """ The center time coordinate of the Gaussian (seconds). """

    anh: float
    """ The anharmonicity of the qubit, f01-f12 (Hertz). """

    alpha: float
    """ Dimensionles DRAG parameter. """

    scale: Optional[float] = None
    """ An optional global scaling factor. """

    phase: Optional[float] = None
    """ An optional phase shift factor. """

    detuning: Optional[float] = None
    """ An optional frequency detuning factor. """

    def out(self) -> str:
        output = "drag_gaussian("
        output += ", ".join(
            [
                f"duration: {self.duration}",
                f"fwhm: {self.fwhm}",
                f"t0: {self.t0}",
                f"anh: {self.anh}",
                f"alpha: {self.alpha}",
            ]
            + _optional_field_strs(self)
        )
        output += ")"
        return output

    def __str__(self) -> str:
        return self.out()

    def samples(self, rate: float) -> np.ndarray:
        ts = np.arange(self.num_samples(rate), dtype=np.complex128) / rate
        sigma = 0.5 * self.fwhm / np.sqrt(2.0 * np.log(2.0))
        env = np.exp(-0.5 * (ts - self.t0) ** 2 / sigma ** 2)
        env_der = (self.alpha * (1.0 / (2 * np.pi * self.anh * sigma ** 2))) * (ts - self.t0) * env
        iqs = env + 1.0j * env_der
        return _update_envelope(iqs, rate, scale=self.scale, phase=self.phase, detuning=self.detuning)


@waveform("hrm_gaussian")
class HrmGaussianWaveform(TemplateWaveform):
    """A Hermite Gaussian waveform.

    REFERENCE: Effects of arbitrary laser or NMR pulse shapes on population
        inversion and coherence Warren S. Warren. 81, (1984); doi:
        10.1063/1.447644
    """

    fwhm: float
    """ The Full-Width-Half-Max of the Gaussian (seconds). """

    t0: float
    """ The center time coordinate of the Gaussian (seconds). """

    anh: float
    """ The anharmonicity of the qubit, f01-f12 (Hertz). """

    alpha: float
    """ Dimensionles DRAG parameter. """

    second_order_hrm_coeff: float
    """ Second order coefficient (see Warren 1984). """

    scale: Optional[float] = None
    """ An optional global scaling factor. """

    phase: Optional[float] = None
    """ An optional phase shift factor. """

    detuning: Optional[float] = None
    """ An optional frequency detuning factor. """

    def out(self) -> str:
        output = "hrm_gaussian("
        output += ", ".join(
            [
                f"duration: {self.duration}",
                f"fwhm: {self.fwhm}",
                f"t0: {self.t0}",
                f"anh: {self.anh}",
                f"alpha: {self.alpha}",
                f"second_order_hrm_coeff: {self.second_order_hrm_coeff}",
            ]
            + _optional_field_strs(self)
        )
        output += ")"
        return output

    def __str__(self) -> str:
        return self.out()

    def samples(self, rate: float) -> np.ndarray:
        ts = np.arange(self.num_samples(rate), dtype=np.complex128) / rate
        sigma = 0.5 * self.fwhm / np.sqrt(2.0 * np.log(2.0))
        exponent_of_t = 0.5 * (ts - self.t0) ** 2 / sigma ** 2
        gauss = np.exp(-exponent_of_t)
        env = (1 - self.second_order_hrm_coeff * exponent_of_t) * gauss
        deriv_prefactor = -self.alpha / (2 * np.pi * self.anh)
        env_der = (
            deriv_prefactor
            * (ts - self.t0)
            / (sigma ** 2)
            * gauss
            * (self.second_order_hrm_coeff * (exponent_of_t - 1) - 1)
        )
        iqs = env + 1.0j * env_der
        return _update_envelope(iqs, rate, scale=self.scale, phase=self.phase, detuning=self.detuning)


@waveform("erf_square")
class ErfSquareWaveform(TemplateWaveform):
    """ A pulse with a flat top and edges that are error functions (erf). """

    risetime: float
    """ The width of each of the rise and fall sections of the pulse (seconds). """
    pad_left: float
    """ Amount of zero-padding to add to the left of the pulse (seconds)."""
    pad_right: float
    """ Amount of zero-padding to add to the right of the pulse (seconds). """

    scale: Optional[float] = None
    """ An optional global scaling factor. """

    phase: Optional[float] = None
    """ An optional phase shift factor. """

    detuning: Optional[float] = None
    """ An optional frequency detuning factor. """

    def out(self) -> str:
        output = "erf_square("
        output += ", ".join(
            [
                f"duration: {self.duration}",
                f"risetime: {self.risetime}",
                f"pad_left: {self.pad_left}",
                f"pad_right: {self.pad_right}",
            ]
            + _optional_field_strs(self)
        )
        output += ")"
        return output

    def __str__(self) -> str:
        return self.out()

    def samples(self, rate: float) -> np.ndarray:
        ts = np.arange(self.num_samples(rate), dtype=np.complex128) / rate
        fwhm = 0.5 * self.risetime
        t1 = fwhm
        t2 = self.duration - fwhm
        sigma = 0.5 * fwhm / np.sqrt(2.0 * np.log(2.0))
        vals = 0.5 * (erf((ts - t1) / sigma) - erf((ts - t2) / sigma))
        zeros_left = np.zeros(int(np.ceil(self.pad_left * rate)), dtype=np.complex128)
        zeros_right = np.zeros(int(np.ceil(self.pad_right * rate)), dtype=np.complex128)
        iqs = np.concatenate((zeros_left, vals, zeros_right))
        return _update_envelope(iqs, rate, scale=self.scale, phase=self.phase, detuning=self.detuning)


@waveform("boxcar_kernel")
class BoxcarAveragerKernel(TemplateWaveform):

    scale: Optional[float] = None
    """ An optional global scaling factor. """

    phase: Optional[float] = None
    """ An optional phase shift factor. """

    detuning: Optional[float] = None
    """ An optional frequency detuning factor. """

    def out(self) -> str:
        output = "boxcar_kernel("
        output += ", ".join([f"duration: {self.duration}"] + _optional_field_strs(self))
        output += ")"
        return output

    def __str__(self) -> str:
        return self.out()

    def samples(self, rate: float) -> np.ndarray:
        n = self.num_samples(rate)
        iqs = np.full(n, 1.0 / n, dtype=np.complex128)
        return _update_envelope(iqs, rate, scale=self.scale, phase=self.phase, detuning=self.detuning)



##############################################################################
# Copyright 2016-2019 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Schema definition of a Experiment, which is a collection of ExperimentSetting
objects and a main program body (or ansatz). This schema is widely useful for defining
and executing many common types of algorithms / applications, including state and process
tomography, and the variational quantum eigensolver.
"""
import json
import logging
import warnings
from json import JSONEncoder
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Union,
    cast,
)

from pyquil.experiment._calibration import CalibrationMethod
from pyquil.experiment._memory import (
    pauli_term_to_measurement_memory_map,
    pauli_term_to_preparation_memory_map,
)
from pyquil.experiment._program import (
    parameterized_single_qubit_measurement_basis,
    parameterized_single_qubit_state_preparation,
    parameterized_readout_symmetrization,
    measure_qubits,
)
from pyquil.experiment._result import ExperimentResult
from pyquil.experiment._setting import ExperimentSetting, _OneQState, TensorProductState
from pyquil.experiment._symmetrization import SymmetrizationLevel
from pyquil.gates import RESET
from pyquil.paulis import PauliTerm, is_identity
from pyquil.quil import Program
from pyquil.quilbase import Reset, ResetQubit


log = logging.getLogger(__name__)


def _abbrev_program(program: Program, max_len: int = 10) -> str:
    """Create an abbreviated string representation of a Program.

    This will join all instructions onto a single line joined by '; '. If the number of
    instructions exceeds ``max_len``, some will be excluded from the string representation.
    """
    program_lines = program.out().splitlines()
    if max_len is not None and len(program_lines) > max_len:
        first_n = max_len // 2
        last_n = max_len - first_n
        excluded = len(program_lines) - max_len
        program_lines = program_lines[:first_n] + [f"... {excluded} instrs not shown ..."] + program_lines[-last_n:]

    return "   " + "\n   ".join(program_lines)


def _remove_reset_from_program(program: Program) -> Program:
    """
    Trim the RESET from a program because in measure_observables it is re-added.

    :param program: Program to remove RESET(s) from.
    :return: Trimmed Program.
    """
    p = program.copy_everything_except_instructions()

    for inst in program:
        if not isinstance(inst, (Reset, ResetQubit)):
            p.inst(inst)

    return p


class Experiment:
    """
    A tomography-like experiment.

    Many near-term quantum algorithms involve:

     - some limited state preparation
     - enacting a quantum process (like in tomography) or preparing a variational ansatz state
       (like in VQE)
     - measuring observables of the state.

    Where we typically use a large number of (state_prep, measure) pairs but keep the ansatz
    program consistent. This class stores the ansatz program as a :py:class:`~pyquil.Program`
    and maintains a list of :py:class:`ExperimentSetting` objects which each represent a
    (state_prep, measure) pair.

    Settings diagonalized by a shared tensor product basis (TPB) can (optionally) be estimated
    simultaneously. Therefore, this class is backed by a list of list of ExperimentSettings.
    Settings sharing an inner list will be estimated simultaneously. If you don't want this,
    provide a list of length-1-lists. As a convenience, if you pass a 1D list to the constructor
    will expand it to a list of length-1-lists.

    This class will not group settings for you. Please see :py:func:`group_experiments` for
    a function that will automatically process a Experiment to group Experiments sharing
    a TPB.

    :ivar settings: The collection of ExperimentSetting objects that define this experiment.
    :ivar program: The main program body of this experiment. Also determines the ``shots``
        and ``reset`` instance variables. The ``shots`` instance variable is the number of
        shots to take per ExperimentSetting. The ``reset`` instance variable is whether to
        actively reset qubits instead of waiting several times the coherence length for qubits
        to decay to ``|0>`` naturally. Setting this to True is much faster but there is a ~1%
        error per qubit in the reset operation. Thermal noise from "traditional" reset is not
        routinely characterized but is of the same order.
    :ivar symmetrization: the level of readout symmetrization to perform for the estimation
        and optional calibration of each observable. The following integer levels, encapsulated in
        the ``SymmetrizationLevel`` integer enum, are currently supported:

        * -1 -- exhaustive symmetrization uses every possible combination of flips
        * 0 -- no symmetrization
        * 1 -- symmetrization using an orthogonal array (OA) with strength 1
        * 2 -- symmetrization using an orthogonal array (OA) with strength 2
        * 3 -- symmetrization using an orthogonal array (OA) with strength 3

        Note that (default) exhaustive symmetrization requires a number of QPU calls exponential in
        the number of qubits in the union of the support of the observables in any group of settings
        in ``tomo_experiment``; the number of shots may need to be increased to accommodate this.
        see :py:func:`run_symmetrized_readout` in api._quantum_computer for more information.
    """

    def __init__(
        self,
        settings: Union[List[ExperimentSetting], List[List[ExperimentSetting]]],
        program: Program,
        *,
        symmetrization: int = SymmetrizationLevel.EXHAUSTIVE,
        calibration: int = CalibrationMethod.PLUS_EIGENSTATE,
    ):
        if len(settings) == 0:
            s: List[List[ExperimentSetting]] = []
        else:
            if isinstance(settings[0], ExperimentSetting):
                # convenience wrapping in lists of length 1
                s = [[expt] for expt in cast(List[ExperimentSetting], settings)]
            else:
                s = cast(List[List[ExperimentSetting]], settings)

        self._settings = s
        self.program = program
        self.symmetrization = SymmetrizationLevel(symmetrization)
        if self.symmetrization != SymmetrizationLevel.EXHAUSTIVE:
            if type(calibration) == int and calibration != 0:
                warnings.warn(
                    "Calibration is only supported for exhaustive symmetrization, "
                    "thus setting self.calibration = 0 (CalibrationMethod.NONE)."
                )
            self.calibration = CalibrationMethod.NONE
        else:
            self.calibration = CalibrationMethod(calibration)
        self.shots = self.program.num_shots

        if "RESET" in self.program.out():
            self.reset = True
            self.program = _remove_reset_from_program(self.program)
        else:
            self.reset = False

    def __len__(self) -> int:
        return len(self._settings)

    def __getitem__(self, item: int) -> List[ExperimentSetting]:
        return self._settings[item]

    def __setitem__(self, key: int, value: List[ExperimentSetting]) -> None:
        self._settings[key] = value

    def __delitem__(self, key: int) -> None:
        self._settings.__delitem__(key)

    def __iter__(self) -> Generator[List[ExperimentSetting], None, None]:
        yield from self._settings

    def __reversed__(self) -> Generator[List[ExperimentSetting], None, None]:
        yield from reversed(self._settings)

    def __contains__(self, item: List[ExperimentSetting]) -> bool:
        return item in self._settings

    def append(self, expts: Union[ExperimentSetting, List[ExperimentSetting]]) -> None:
        if not isinstance(expts, list):
            expts = [expts]
        self._settings.append(expts)

    def count(self, expt: List[ExperimentSetting]) -> int:
        return self._settings.count(expt)

    def index(self, expt: List[ExperimentSetting], start: int = 0, stop: int = 0) -> int:
        return self._settings.index(expt, start, stop)

    def extend(self, expts: List[List[ExperimentSetting]]) -> None:
        self._settings.extend(expts)

    def insert(self, index: int, expt: List[ExperimentSetting]) -> None:
        self._settings.insert(index, expt)

    def pop(self, index: int = 0) -> List[ExperimentSetting]:
        return self._settings.pop(index)

    def remove(self, expt: List[ExperimentSetting]) -> None:
        self._settings.remove(expt)

    def reverse(self) -> None:
        self._settings.reverse()

    def sort(self, key: Optional[Callable[[List[ExperimentSetting]], Any]] = None, reverse: bool = False) -> None:
        return self._settings.sort(key=key, reverse=reverse)

    def setting_strings(self) -> Generator[str, None, None]:
        yield from (
            "{i}: {st_str}".format(i=i, st_str=", ".join(str(setting) for setting in settings))
            for i, settings in enumerate(self._settings)
        )

    def settings_string(self, abbrev_after: Optional[int] = None) -> str:
        setting_strs = list(self.setting_strings())
        if abbrev_after is not None and len(setting_strs) > abbrev_after:
            first_n = abbrev_after // 2
            last_n = abbrev_after - first_n
            excluded = len(setting_strs) - abbrev_after
            setting_strs = setting_strs[:first_n] + [f"... {excluded} settings not shown ..."] + setting_strs[-last_n:]
        return "   " + "\n   ".join(setting_strs)

    def __repr__(self) -> str:
        string = f"shots: {self.shots}\n"
        if self.reset:
            string += "active reset: enabled\n"
        else:
            string += "active reset: disabled\n"
        string += f"symmetrization: {self.symmetrization} ({self.symmetrization.name.lower()})\n"
        string += f"calibration: {self.calibration} ({self.calibration.name.lower()})\n"
        string += f"program:\n{_abbrev_program(self.program)}\n"
        string += f"settings:\n{self.settings_string(abbrev_after=20)}"
        return string

    def serializable(self) -> Dict[str, Any]:
        return {
            "type": "Experiment",
            "settings": self._settings,
            "program": self.program.out(),
            "symmetrization": self.symmetrization,
            "shots": self.shots,
            "reset": self.reset,
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Experiment):
            return False
        return self.serializable() == other.serializable()

    def get_meas_qubits(self) -> List[int]:
        """
        Return the sorted list of qubits that are involved in the all the out_operators of the
        settings for this ``Experiment`` object.
        """
        meas_qubits: Set[int] = set()
        for settings in self:
            assert len(settings) == 1
            meas_qubits.update(cast(List[int], settings[0].out_operator.get_qubits()))
        return sorted(meas_qubits)

    def get_meas_registers(self, qubits: Optional[Sequence[int]] = None) -> List[int]:
        """
        Return the sorted list of memory registers corresponding to the list of qubits provided.
        If no qubits are provided, just returns the list of numbers from 0 to n-1 where n is the
        number of qubits resulting from the ``get_meas_qubits`` method.
        """
        meas_qubits = self.get_meas_qubits()

        if qubits is None:
            return list(range(len(meas_qubits)))

        meas_registers = []
        for q in qubits:
            meas_registers.append(meas_qubits.index(q))
        return sorted(meas_registers)

    def generate_experiment_program(self) -> Program:
        """
        Generate a parameterized program containing the main body program along with some additions
        to support the various state preparation, measurement, and symmetrization specifications of
        this ``Experiment``.

        State preparation and measurement are achieved via ZXZXZ-decomposed single-qubit gates,
        where the angles of each ``RZ`` rotation are declared parameters that can be assigned at
        runtime. Symmetrization is achieved by putting an ``RX`` gate (also parameterized by a
        declared value) before each ``MEASURE`` operation. In addition, a ``RESET`` operation
        is prepended to the ``Program`` if the experiment has active qubit reset enabled. Finally,
        each qubit specified in the settings is measured, and the number of shots is added.

        :return: Parameterized ``Program`` that is capable of collecting statistics for every
            ``ExperimentSetting`` in this ``Experiment``.
        """
        meas_qubits = self.get_meas_qubits()

        p = Program()

        if self.reset:
            if any(isinstance(instr, (Reset, ResetQubit)) for instr in self.program):
                raise ValueError("RESET already added to program")
            p += RESET()

        for settings in self:
            assert len(settings) == 1
            if ("X" in str(settings[0].in_state)) or ("Y" in str(settings[0].in_state)):
                if "DECLARE preparation_alpha" in self.program.out():
                    raise ValueError('Memory "preparation_alpha" has been declared already.')
                if "DECLARE preparation_beta" in self.program.out():
                    raise ValueError('Memory "preparation_beta" has been declared already.')
                if "DECLARE preparation_gamma" in self.program.out():
                    raise ValueError('Memory "preparation_gamma" has been declared already.')
                p += parameterized_single_qubit_state_preparation(meas_qubits)
                break

        p += self.program

        for settings in self:
            assert len(settings) == 1
            if ("X" in str(settings[0].out_operator)) or ("Y" in str(settings[0].out_operator)):
                if "DECLARE measurement_alpha" in self.program.out():
                    raise ValueError('Memory "measurement_alpha" has been declared already.')
                if "DECLARE measurement_beta" in self.program.out():
                    raise ValueError('Memory "measurement_beta" has been declared already.')
                if "DECLARE measurement_gamma" in self.program.out():
                    raise ValueError('Memory "measurement_gamma" has been declared already.')
                p += parameterized_single_qubit_measurement_basis(meas_qubits)
                break

        if self.symmetrization != 0:
            if "DECLARE symmetrization" in self.program.out():
                raise ValueError('Memory "symmetrization" has been declared already.')
            p += parameterized_readout_symmetrization(meas_qubits)

        if "DECLARE ro" in self.program.out():
            raise ValueError('Memory "ro" has already been declared for this program.')
        p += measure_qubits(meas_qubits)

        p.wrap_in_numshots_loop(self.shots)

        return p

    def build_setting_memory_map(self, setting: ExperimentSetting) -> Dict[str, List[float]]:
        """
        Build the memory map corresponding to the state preparation and measurement specifications
        encoded in the provided ``ExperimentSetting``, taking into account the full set of qubits
        that are present in the ``Experiment`` object.

        :return: Memory map for state prep and measurement.
        """
        meas_qubits = self.get_meas_qubits()

        in_pt = PauliTerm.from_list([(op, meas_qubits.index(cast(int, q))) for q, op in setting._in_operator()])
        out_pt = PauliTerm.from_list([(op, meas_qubits.index(cast(int, q))) for q, op in setting.out_operator])

        preparation_map = pauli_term_to_preparation_memory_map(in_pt)
        measurement_map = pauli_term_to_measurement_memory_map(out_pt)

        return {**preparation_map, **measurement_map}

    def build_symmetrization_memory_maps(
        self, qubits: Sequence[int], label: str = "symmetrization"
    ) -> List[Dict[str, List[float]]]:
        """
        Build a list of memory maps to be used in a program that is trying to perform readout
        symmetrization via parametric compilation. For example, if we have the following program:

            RX(symmetrization[0]) 0
            RX(symmetrization[1]) 1
            MEASURE 0 ro[0]
            MEASURE 1 ro[1]

        We can perform exhaustive readout symmetrization on our two qubits by providing the four
        following memory maps, and then appropriately flipping the resultant bitstrings:

            {'symmetrization': [0.0, 0.0]} -> XOR results with [0,0]
            {'symmetrization': [0.0, pi]}  -> XOR results with [0,1]
            {'symmetrization': [pi, 0.0]}  -> XOR results with [1,0]
            {'symmetrization': [pi, pi]}   -> XOR results with [1,1]

        :param qubits: List of qubits to symmetrize readout for.
        :param label: Name of the declared memory region. Defaults to "symmetrization".
        :return: List of memory maps that performs the desired level of symmetrization.
        """
        num_meas_registers = len(self.get_meas_qubits())
        symm_registers = self.get_meas_registers(qubits)

        if self.symmetrization == SymmetrizationLevel.NONE:
            return [{}]

        # TODO: add support for orthogonal arrays
        if self.symmetrization != SymmetrizationLevel.EXHAUSTIVE:
            raise ValueError("We only support exhaustive symmetrization for now.")

        import numpy as np
        import itertools

        assignments = itertools.product(np.array([0, np.pi]), repeat=len(symm_registers))
        memory_maps = []
        for a in assignments:
            zeros = np.zeros(num_meas_registers)
            for idx, r in enumerate(symm_registers):
                zeros[r] = a[idx]
            memory_maps.append({f"{label}": list(zeros)})
        return memory_maps

    def generate_calibration_experiment(self) -> "Experiment":
        """
        Generate another ``Experiment`` object that can be used to calibrate the various multi-qubit
        observables involved in this ``Experiment``. This is achieved by preparing the plus-one
        (minus-one) eigenstate of each ``out_operator``, and measuring the resulting expectation
        value of the same ``out_operator``. Ideally, this would always give +1 (-1), but when
        symmetric readout error is present the effect is to scale the resultant expectations by some
        constant factor. Determining this scale factor is what we call *readout calibration*, and
        then the readout error in subsequent measurements can then be mitigated by simply dividing
        by the scale factor.

        :return: A new ``Experiment`` that can calibrate the readout error of all the
            observables involved in this experiment.
        """
        if self.calibration != CalibrationMethod.PLUS_EIGENSTATE:
            raise ValueError('We currently only support the "plus eigenstate" calibration method.')

        calibration_settings = []
        for settings in self:
            assert len(settings) == 1
            calibration_settings.append(
                ExperimentSetting(
                    in_state=_pauli_to_product_state(settings[0].out_operator),
                    out_operator=settings[0].out_operator,
                    additional_expectations=settings[0].additional_expectations,
                )
            )

        calibration_program = Program()
        if self.reset:
            calibration_program += RESET()
        calibration_program.wrap_in_numshots_loop(self.shots)

        if self.symmetrization != SymmetrizationLevel.EXHAUSTIVE:
            raise ValueError("We currently only support calibration for exhaustive symmetrization")

        return Experiment(
            settings=calibration_settings,
            program=calibration_program,
            symmetrization=SymmetrizationLevel.EXHAUSTIVE,
            calibration=CalibrationMethod.NONE,
        )


def _pauli_to_product_state(in_state: PauliTerm) -> TensorProductState:
    """
    Convert a Pauli term to a TensorProductState.
    """
    if is_identity(in_state):
        return TensorProductState()
    else:
        return TensorProductState(
            [
                _OneQState(label=pauli_label, index=0, qubit=cast(int, qubit))
                for qubit, pauli_label in in_state._ops.items()
            ]
        )


class OperatorEncoder(JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, ExperimentSetting):
            return o.serializable()
        if isinstance(o, Experiment):
            return o.serializable()
        if isinstance(o, ExperimentResult):
            return o.serializable()
        return o


def to_json(fn: str, obj: Any) -> str:
    """
    Convenience method to save pyquil.experiment objects as a JSON file.

    See :py:func:`read_json`.
    """
    # Specify UTF-8 to guard against systems that default to an ASCII locale.
    with open(fn, "w", encoding="utf-8") as f:
        json.dump(obj, f, cls=OperatorEncoder, indent=2, ensure_ascii=False)
    return fn


def _operator_object_hook(obj: Mapping[str, Any]) -> Union[Mapping[str, Any], Experiment]:
    if "type" in obj and obj["type"] in ["Experiment", "TomographyExperiment"]:
        # I bet this doesn't work for grouped experiment settings
        settings = [[ExperimentSetting.from_str(s) for s in stt] for stt in obj["settings"]]
        p = Program(obj["program"])
        p.wrap_in_numshots_loop(obj["shots"])
        ex = Experiment(settings=settings, program=p, symmetrization=obj["symmetrization"])
        ex.reset = obj["reset"]
        return ex
    return obj


def read_json(fn: str) -> Any:
    """
    Convenience method to read pyquil.experiment objects from a JSON file.

    See :py:func:`to_json`.
    """
    # Specify UTF-8 to guard against systems that default to an ASCII locale.
    with open(fn, encoding="utf-8") as f:
        return json.load(f, object_hook=_operator_object_hook)



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Generic, Mapping, Optional, TypeVar

import numpy as np
from pyquil.api._abstract_compiler import QuantumExecutable


class QAMError(RuntimeError):
    pass


T = TypeVar("T")
"""A generic parameter describing the opaque job handle returned from QAM#execute and subclasses."""


@dataclass
class QAMExecutionResult:
    executable: QuantumExecutable
    """The executable corresponding to this result."""

    readout_data: Mapping[str, Optional[np.ndarray]] = field(default_factory=dict)
    """Readout data returned from the QAM, keyed on the name of the readout register or post-processing node."""

    execution_duration_microseconds: Optional[int] = field(default=None)
    """Duration job held exclusive hardware access. Defaults to ``None`` when information is not available."""


class QAM(ABC, Generic[T]):
    """
    Quantum Abstract Machine: This class acts as a generic interface describing how a classical
    computer interacts with a live quantum computer.
    """

    @abstractmethod
    def execute(self, executable: QuantumExecutable) -> T:
        """
        Run an executable on a QAM, returning a handle to be used to retrieve
        results.

        :param executable: The executable program to be executed by the QAM.
        """

    @abstractmethod
    def get_result(self, execute_response: T) -> QAMExecutionResult:
        """
        Retrieve the results associated with a previous call to ``QAM#execute``.

        :param execute_response: The return value from a call to ``execute``.
        """

    def run(self, executable: QuantumExecutable) -> QAMExecutionResult:
        """
        Run an executable to completion on the QAM.
        """
        return self.get_result(self.execute(executable))



##############################################################################
# Copyright 2016-2021 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Optional, cast, Tuple, Union, List, Any
from attr import field

import rpcq
from dateutil.parser import parse as parsedate
from dateutil.tz import tzutc
from qcs_api_client.models import EngagementWithCredentials, EngagementCredentials
from retry import retry

from pyquil.api import EngagementManager


@dataclass
class RunProgramRequest:
    """
    Request to run a program.
    """

    id: str
    """Identifier for request."""

    priority: int
    """Priority for request."""

    program: str
    """Encrypted program to run."""

    patch_values: Dict[str, List[Union[int, float]]]
    """Map of data names to data values for patching the program."""


@dataclass
class RunProgramResponse:
    """
    Run program response.
    """

    job_id: str
    """Identifier for created run job."""


@dataclass
class GetBuffersRequest:
    """
    Request for getting job buffers.
    """

    job_id: str
    """Job for which to get buffers."""

    wait: bool
    """Whether or not to wait until buffers become available."""


@dataclass
class BufferResponse:
    """
    Job buffer.
    """

    shape: Tuple[int, int]
    """Shape of the buffer: (<trials>, <slots>)."""

    dtype: str
    """Buffer data type."""

    data: bytes
    """Raw buffer data (C order)."""


@dataclass
class GetBuffersResponse:
    """
    Job buffers response.
    """

    buffers: Dict[str, BufferResponse]
    """Job buffers, by buffer name."""

    execution_duration_microseconds: Optional[int] = field(default=None)
    "Duration job held exclusive hardware access."


class QPUClient:
    """
    Client for making requests to a QPU.
    """

    def __init__(
        self,
        *,
        quantum_processor_id: str,
        engagement_manager: EngagementManager,
        endpoint_id: Optional[str] = None,
        request_timeout: float = 10.0,
    ) -> None:
        """
        Instantiate a new QPU client, authenticated using the given engagement.

        :param quantum_processor_id: ID of quantum processor to target.
        :param engagement_manager: Manager for QPU engagements.
        :param request_timeout: Timeout for requests, in seconds.
        """
        self.quantum_processor_id = quantum_processor_id
        self._endpoint_id = endpoint_id
        self._engagement_manager = engagement_manager
        self.timeout = request_timeout

    def run_program(self, request: RunProgramRequest) -> RunProgramResponse:
        """
        Run a program on a QPU.
        """
        rpcq_request = rpcq.messages.QPURequest(
            id=request.id,
            program=request.program,
            patch_values=request.patch_values,
        )
        job_id = self._rpcq_request(
            "execute_qpu_request",
            request=rpcq_request,
            priority=request.priority,
            user=None,
        )
        return RunProgramResponse(job_id=job_id)

    def get_buffers(self, request: GetBuffersRequest) -> GetBuffersResponse:
        """
        Get job buffers.
        """
        buffs = self._rpcq_request(
            "get_buffers",
            job_id=request.job_id,
            wait=request.wait,
        )
        return GetBuffersResponse(
            buffers={
                name: BufferResponse(
                    shape=cast(Tuple[int, int], tuple(val["shape"])),
                    dtype=val["dtype"],
                    data=val["data"],
                )
                for name, val in buffs.items()
            }
        )

    def get_execution_results(self, request: GetBuffersRequest) -> GetBuffersResponse:
        """
        Get job buffers and execution metadata.
        """
        result = self._rpcq_request(
            "get_execution_results",
            job_id=request.job_id,
            wait=request.wait,
        )
        return GetBuffersResponse(
            buffers={
                name: BufferResponse(
                    shape=cast(Tuple[int, int], tuple(val["shape"])),
                    dtype=val["dtype"],
                    data=val["data"],
                )
                for name, val in result.buffers.items()
            },
            execution_duration_microseconds=result.execution_duration_microseconds,
        )

    @retry(exceptions=TimeoutError, tries=2)  # type: ignore
    def _rpcq_request(self, method_name: str, *args: Any, **kwargs: Any) -> Any:
        engagement = self._engagement_manager.get_engagement(
            endpoint_id=self._endpoint_id,
            quantum_processor_id=self.quantum_processor_id,
            request_timeout=self.timeout,
        )
        client = rpcq.Client(
            endpoint=engagement.address,
            timeout=self._calculate_timeout(engagement),
            auth_config=self._auth_config(engagement.credentials),
        )
        try:
            return client.call(method_name, *args, **kwargs)
        finally:
            client.close()  # type: ignore

    def _calculate_timeout(self, engagement: EngagementWithCredentials) -> float:
        engagement_time_left = parsedate(engagement.expires_at) - datetime.now(tzutc())
        return min(self.timeout, engagement_time_left.total_seconds())

    @staticmethod
    def _auth_config(credentials: EngagementCredentials) -> rpcq.ClientAuthConfig:
        return rpcq.ClientAuthConfig(
            client_secret_key=credentials.client_secret.encode(),
            client_public_key=credentials.client_public.encode(),
            server_public_key=credentials.server_public.encode(),
        )
